@@ -47,7 +47,8 @@ const hello_ext_entry_st ext_mod_key_share = {
 	.name = "Key Share",
 	.tls_id = 51,
 	.gid = GNUTLS_EXTENSION_KEY_SHARE,
-	.parse_type = _GNUTLS_EXT_TLS_POST_CS,
+	.client_parse_point = _GNUTLS_EXT_TLS_POST_CS,
+	.server_parse_point = _GNUTLS_EXT_TLS_POST_CS,
 	.validity = GNUTLS_EXT_FLAG_TLS | GNUTLS_EXT_FLAG_CLIENT_HELLO | GNUTLS_EXT_FLAG_TLS13_SERVER_HELLO |
 		    GNUTLS_EXT_FLAG_HRR,
 	.recv_func = key_share_recv_params,
@@ -504,11 +505,10 @@ client_use_key_share(gnutls_session_t session, const gnutls_group_entry_st *grou
 
 static int
 key_share_recv_params(gnutls_session_t session,
-		      const uint8_t * data, size_t _data_size)
+		      const uint8_t * data, size_t data_size)
 {
 	int ret;
-	ssize_t data_size = _data_size;
-	ssize_t size;
+	size_t size;
 	unsigned gid;
 	const version_entry_st *ver;
 	const gnutls_group_entry_st *group;
