@@ -96,14 +96,6 @@ typedef int ssize_t;
 # define FALLTHROUGH
 #endif
 
-
-/* some systems had problems with long long int, thus,
- * it is not used.
- */
-typedef struct {
-	unsigned char i[8];
-} gnutls_uint64;
-
 #include <gnutls/gnutls.h>
 #include <gnutls/dtls.h>
 #include <gnutls/abstract.h>
@@ -256,24 +248,21 @@ typedef enum record_send_state_t {
 
 #define MEMSUB(x,y) ((ssize_t)((ptrdiff_t)x-(ptrdiff_t)y))
 
-#define DECR_LEN(len, x) do { len-=x; if (len<0) {gnutls_assert(); return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;} } while (0)
+#define DECR_LEN(len, x) DECR_LENGTH_RET(len, x, GNUTLS_E_UNEXPECTED_PACKET_LENGTH)
 #define DECR_LEN_FINAL(len, x) do { \
-	len-=x; \
-	if (len != 0) \
+	if (len != x) \
 		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET_LENGTH); \
+	else \
+		len = 0; \
 	} while (0)
-#define DECR_LENGTH_RET(len, x, RET) do { len-=x; if (len<0) {gnutls_assert(); return RET;} } while (0)
-#define DECR_LENGTH_COM(len, x, COM) do { len-=x; if (len<0) {gnutls_assert(); COM;} } while (0)
+#define DECR_LENGTH_RET(len, x, RET) DECR_LENGTH_COM(len, x, return RET)
+#define DECR_LENGTH_COM(len, x, COM) do { if (len<x) {gnutls_assert(); COM;} else len-=x; } while (0)
 
 #define GNUTLS_POINTER_TO_INT(_) ((int) GNUTLS_POINTER_TO_INT_CAST (_))
 #define GNUTLS_INT_TO_POINTER(_) ((void*) GNUTLS_POINTER_TO_INT_CAST (_))
 
 #define GNUTLS_KX_INVALID (-1)
 
-typedef struct {
-	uint8_t pint[3];
-} uint24;
-
 #include <mpi.h>
 
 typedef enum handshake_state_t { STATE0 = 0, STATE1, STATE2,
@@ -461,7 +450,7 @@ typedef struct mbuffer_st {
 	content_type_t type;
 
 	/* record layer sequence */
-	gnutls_uint64 record_sequence;
+	uint64_t record_sequence;
 
 	/* Filled in by handshake layer on send:
 	 * type, epoch, htype, handshake_sequence
@@ -644,6 +633,10 @@ typedef struct record_state_st record_state_st;
 struct record_parameters_st;
 typedef struct record_parameters_st record_parameters_st;
 
+#define GNUTLS_CIPHER_FLAG_ONLY_AEAD	(1 << 0) /* When set, this cipher is only available through the new AEAD API */
+#define GNUTLS_CIPHER_FLAG_XOR_NONCE	(1 << 1) /* In this TLS AEAD cipher xor the implicit_iv with the nonce */
+#define GNUTLS_CIPHER_FLAG_NO_REKEY	(1 << 2) /* whether this tls1.3 cipher doesn't need to rekey after 2^24 messages */
+
 /* cipher and mac parameters */
 typedef struct cipher_entry_st {
 	const char *name;
@@ -655,9 +648,7 @@ typedef struct cipher_entry_st {
 	uint16_t explicit_iv;	/* the size of explicit IV - the IV stored in record */
 	uint16_t cipher_iv;	/* the size of IV needed by the cipher */
 	uint16_t tagsize;
-	bool xor_nonce;	/* In this TLS AEAD cipher xor the implicit_iv with the nonce */
-	bool only_aead; /* When set, this cipher is only available through the new AEAD API */
-	bool no_rekey; /* whether this tls1.3 cipher doesn't need to rekey after 2^24 messages */
+	unsigned flags;
 } cipher_entry_st;
 
 typedef struct gnutls_cipher_suite_entry_st {
@@ -690,6 +681,8 @@ typedef struct gnutls_group_entry_st {
 	unsigned tls_id;		/* The RFC4492 namedCurve ID or TLS 1.3 group ID */
 } gnutls_group_entry_st;
 
+#define GNUTLS_MAC_FLAG_PREIMAGE_INSECURE	1  /* if this algorithm should not be trusted for pre-image attacks */
+#define GNUTLS_MAC_FLAG_CONTINUOUS_MAC		(1 << 1) /* if this MAC should be used in a 'continuous' way in TLS */
 /* This structure is used both for MACs and digests
  */
 typedef struct mac_entry_st {
@@ -702,7 +695,7 @@ typedef struct mac_entry_st {
 	unsigned nonce_size;
 	unsigned placeholder;	/* if set, then not a real MAC */
 	unsigned block_size;	/* internal block size for HMAC */
-	unsigned preimage_insecure; /* if this algorithm should not be trusted for pre-image attacks */
+	unsigned flags;
 } mac_entry_st;
 
 typedef struct {
@@ -872,7 +865,7 @@ struct record_state_st {
 	} ctx;
 	unsigned aead_tag_size;
 	unsigned is_aead;
-	gnutls_uint64 sequence_number;
+	uint64_t sequence_number;
 };
 
 
