@@ -17,12 +17,12 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
-#ifndef GNUTLS_EXTENSIONS_H
-#define GNUTLS_EXTENSIONS_H
+#ifndef GNUTLS_LIB_HELLO_EXT_H
+#define GNUTLS_LIB_HELLO_EXT_H
 
 #include "gnutls_int.h"
 #include <gnutls/gnutls.h>
@@ -121,7 +121,8 @@ typedef struct hello_ext_entry_st {
 	uint16_t tls_id;
 	unsigned gid; /* gnutls internal ID */
 
-	gnutls_ext_parse_type_t parse_type;
+	gnutls_ext_parse_type_t client_parse_point;
+	gnutls_ext_parse_type_t server_parse_point;
 	unsigned validity; /* multiple items of gnutls_ext_flags_t */
 
 	/* this function must return 0 when Not Applicable
@@ -159,7 +160,7 @@ typedef struct hello_ext_entry_st {
 inline static unsigned
 _gnutls_hello_ext_is_present(gnutls_session_t session, extensions_t id)
 {
-	if (id != 0 && ((1 << id) & session->internals.used_exts))
+	if ((1 << id) & session->internals.used_exts)
 		return 1;
 
 	return 0;
@@ -194,4 +195,4 @@ void _gnutls_hello_ext_save_sr(gnutls_session_t session)
 	_gnutls_hello_ext_save(session, GNUTLS_EXTENSION_SAFE_RENEGOTIATION, 1);
 }
 
-#endif
+#endif /* GNUTLS_LIB_HELLO_EXT_H */
