@@ -17,7 +17,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
@@ -361,8 +361,8 @@ static int append_response(gnutls_certificate_credentials_t sc, unsigned idx,
  * If the response is already expired at the time of loading the code
  * %GNUTLS_E_EXPIRED is returned.
  *
- * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned,
- *   otherwise a negative error code is returned.
+ * Returns: On success, the number of loaded responses is returned,
+ *   otherwise a negative error code.
  *
  * Since: 3.1.3
  **/
@@ -473,7 +473,6 @@ gnutls_certificate_set_ocsp_status_request_mem(gnutls_certificate_credentials_t
 			nresp++;
 
 			gnutls_free(der.data);
-			der.data = NULL;
 
 			p.data++;
 			p.size--;
@@ -609,6 +608,10 @@ gnutls_certificate_get_ocsp_expiration(gnutls_certificate_credentials_t sc,
  * explicit OCSP validity check on the peer's certificate. Should be called after
  * any of gnutls_certificate_verify_peers*() are called.
  *
+ * This function is always usable on client side, but on server side only
+ * under TLS 1.3, which is the first version of TLS that allows cliend-side OCSP
+ * responses.
+ *
  * Returns: non zero if the response was valid, or a zero if it wasn't sent,
  * or sent and was invalid.
  *
