@@ -15,13 +15,13 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
 
 /* Work around problem reported in
-   <http://permalink.gmane.org/gmane.comp.lib.gnulib.bugs/15755>.*/
+   <https://permalink.gmane.org/gmane.comp.lib.gnulib.bugs/15755>.*/
 #if GETTIMEOFDAY_CLOBBERS_LOCALTIME
 #undef localtime
 #endif
@@ -52,6 +52,7 @@
 
 const char str_unknown[] = "(unknown)";
 
+static FILE *logfile = NULL;
 /* Hex encodes the given data adding a semicolon between hex bytes.
  */
 const char *raw_to_string(const unsigned char *raw, size_t raw_size)
@@ -113,45 +114,6 @@ const char *raw_to_base64(const unsigned char *raw, size_t raw_size)
 	return buf;
 }
 
-static void print_x509_info_compact(gnutls_session_t session, int print_crt_status)
-{
-	gnutls_x509_crt_t crt;
-	const gnutls_datum_t *cert_list;
-	unsigned int cert_list_size = 0;
-	int ret;
-	gnutls_datum_t cinfo;
-
-	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
-	if (cert_list_size == 0) {
-		if (print_crt_status)
-			fprintf(stderr, "No certificates found!\n");
-		return;
-	}
-
-	ret = gnutls_x509_crt_init(&crt);
-	if (ret < 0) {
-		fprintf(stderr, "Memory error\n");
-		return;
-	}
-
-	ret =
-	    gnutls_x509_crt_import(crt, &cert_list[0],
-				   GNUTLS_X509_FMT_DER);
-	if (ret < 0) {
-		fprintf(stderr, "Decoding error: %s\n",
-			gnutls_strerror(ret));
-		return;
-	}
-
-	ret = gnutls_x509_crt_print(crt, GNUTLS_CRT_PRINT_COMPACT, &cinfo);
-	if (ret == 0) {
-		printf("- X.509 cert: %s\n", cinfo.data);
-		gnutls_free(cinfo.data);
-	}
-
-	gnutls_x509_crt_deinit(crt);
-}
-
 static void
 print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert, int print_crt_status)
 {
@@ -167,8 +129,8 @@ print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert, i
 		return;
 	}
 
-	fprintf(out, "- Certificate type: X.509\n");
-	fprintf(out, "- Got a certificate list of %d certificates.\n",
+	log_msg(out, "- Certificate type: X.509\n");
+	log_msg(out, "- Got a certificate list of %d certificates.\n",
 	       cert_list_size);
 
 	for (j = 0; j < cert_list_size; j++) {
@@ -189,13 +151,13 @@ print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert, i
 			return;
 		}
 
-		fprintf(out, "- Certificate[%d] info:\n - ", j);
+		log_msg(out, "- Certificate[%d] info:\n - ", j);
 		if (flag == GNUTLS_CRT_PRINT_COMPACT && j > 0)
 			flag = GNUTLS_CRT_PRINT_ONELINE;
 
 		ret = gnutls_x509_crt_print(crt, flag, &cinfo);
 		if (ret == 0) {
-			fprintf(out, "%s\n", cinfo.data);
+			log_msg(out, "%s\n", cinfo.data);
 			gnutls_free(cinfo.data);
 		}
 
@@ -211,9 +173,7 @@ print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert, i
 				return;
 			}
 
-			fputs("\n", out);
-			fputs((char*)pem.data, out);
-			fputs("\n", out);
+			log_msg(out, "\n%s\n", (char*)pem.data);
 
 			gnutls_free(pem.data);
 		}
@@ -222,6 +182,56 @@ print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert, i
 	}
 }
 
+static void
+print_rawpk_info(gnutls_session_t session, FILE *out, int flag, int print_cert, int print_crt_status)
+{
+	gnutls_pcert_st pk_cert;
+	gnutls_pk_algorithm_t pk_algo;
+	const gnutls_datum_t *cert_list;
+	unsigned int cert_list_size = 0;
+	int ret;
+
+	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
+	if (cert_list_size == 0) {
+		if (print_crt_status)
+			fprintf(stderr, "No certificates found!\n");
+		return;
+	}
+
+	log_msg(out, "- Certificate type: Raw Public Key\n");
+	log_msg(out, "- Got %d Raw public-key(s).\n",
+	       cert_list_size);
+
+
+	ret = gnutls_pcert_import_rawpk_raw(&pk_cert, cert_list, GNUTLS_X509_FMT_DER, 0, 0);
+	if (ret < 0) {
+		fprintf(stderr, "Decoding error: %s\n",
+			gnutls_strerror(ret));
+		return;
+	}
+
+	pk_algo = gnutls_pubkey_get_pk_algorithm(pk_cert.pubkey, NULL);
+
+	log_msg(out, "- Raw pk info:\n");
+	log_msg(out, " - PK algo: %s\n", gnutls_pk_algorithm_get_name(pk_algo));
+
+	if (print_cert) {
+		gnutls_datum_t pem;
+
+		ret = gnutls_pubkey_export2(pk_cert.pubkey, GNUTLS_X509_FMT_PEM, &pem);
+		if (ret < 0) {
+			fprintf(stderr, "Encoding error: %s\n",
+				gnutls_strerror(ret));
+			return;
+		}
+
+		log_msg(out, "\n%s\n", (char*)pem.data);
+
+		gnutls_free(pem.data);
+	}
+
+}
+
 /* returns false (0) if not verified, or true (1) otherwise 
  */
 int cert_verify(gnutls_session_t session, const char *hostname, const char *purpose)
@@ -249,12 +259,12 @@ int cert_verify(gnutls_session_t session, const char *hostname, const char *purp
 
 	rc = gnutls_certificate_verify_peers(session, data, elements, &status);
 	if (rc == GNUTLS_E_NO_CERTIFICATE_FOUND) {
-		printf("- Peer did not send any certificate.\n");
+		log_msg(stdout, "- Peer did not send any certificate.\n");
 		return 0;
 	}
 
 	if (rc < 0) {
-		printf("- Could not verify certificate (err: %s)\n",
+		log_msg(stdout, "- Could not verify certificate (err: %s)\n",
 		       gnutls_strerror(rc));
 		return 0;
 	}
@@ -263,12 +273,12 @@ int cert_verify(gnutls_session_t session, const char *hostname, const char *purp
 	rc = gnutls_certificate_verification_status_print(status, type,
 							  &out, 0);
 	if (rc < 0) {
-		printf("- Could not print verification flags (err: %s)\n",
+		log_msg(stdout, "- Could not print verification flags (err: %s)\n",
 		       gnutls_strerror(rc));
 		return 0;
 	}
 
-	printf("- Status: %s\n", out.data);
+	log_msg(stdout, "- Status: %s\n", out.data);
 
 	gnutls_free(out.data);
 
@@ -283,98 +293,101 @@ print_dh_info(gnutls_session_t session, const char *str, int print)
 {
 #if defined(ENABLE_DHE) || defined(ENABLE_ANON)
 	unsigned group;
+	int ret;
+	gnutls_datum_t raw_gen = { NULL, 0 };
+	gnutls_datum_t raw_prime = { NULL, 0 };
+	gnutls_dh_params_t dh_params = NULL;
+	unsigned char *params_data = NULL;
+	size_t params_data_size = 0;
+
+	if (!print)
+		return;
 
 	group = gnutls_group_get(session);
 	if (group != 0) {
 		return;
 	}
 
-	printf("- %sDiffie-Hellman parameters\n", str);
-	printf(" - Using prime: %d bits\n",
+	log_msg(stdout, "- %sDiffie-Hellman parameters\n", str);
+	log_msg(stdout, " - Using prime: %d bits\n",
 	       gnutls_dh_get_prime_bits(session));
-	printf(" - Secret key: %d bits\n",
+	log_msg(stdout, " - Secret key: %d bits\n",
 	       gnutls_dh_get_secret_bits(session));
-	printf(" - Peer's public key: %d bits\n",
+	log_msg(stdout, " - Peer's public key: %d bits\n",
 	       gnutls_dh_get_peers_public_bits(session));
 
-	if (print) {
-		int ret;
-		gnutls_datum_t raw_gen = { NULL, 0 };
-		gnutls_datum_t raw_prime = { NULL, 0 };
-		gnutls_dh_params_t dh_params = NULL;
-		unsigned char *params_data = NULL;
-		size_t params_data_size = 0;
-
-		ret = gnutls_dh_get_group(session, &raw_gen, &raw_prime);
-		if (ret) {
-			fprintf(stderr, "gnutls_dh_get_group %d\n", ret);
-			goto out;
-		}
-
-		ret = gnutls_dh_params_init(&dh_params);
-		if (ret) {
-			fprintf(stderr, "gnutls_dh_params_init %d\n", ret);
-			goto out;
-		}
-
-		ret =
-		    gnutls_dh_params_import_raw(dh_params, &raw_prime,
-						&raw_gen);
-		if (ret) {
-			fprintf(stderr, "gnutls_dh_params_import_raw %d\n",
-				ret);
-			goto out;
-		}
-
-		ret = gnutls_dh_params_export_pkcs3(dh_params,
-						    GNUTLS_X509_FMT_PEM,
-						    params_data,
-						    &params_data_size);
-		if (ret != GNUTLS_E_SHORT_MEMORY_BUFFER) {
-			fprintf(stderr,
-				"gnutls_dh_params_export_pkcs3 %d\n", ret);
-			goto out;
-		}
-
-		params_data = gnutls_malloc(params_data_size);
-		if (!params_data) {
-			fprintf(stderr, "gnutls_malloc %d\n", ret);
-			goto out;
-		}
-
-		ret = gnutls_dh_params_export_pkcs3(dh_params,
-						    GNUTLS_X509_FMT_PEM,
-						    params_data,
-						    &params_data_size);
-		if (ret) {
-			fprintf(stderr,
-				"gnutls_dh_params_export_pkcs3-2 %d\n",
-				ret);
-			goto out;
-		}
-
-		printf(" - PKCS#3 format:\n\n%.*s\n",
-		       (int) params_data_size, params_data);
-
-	      out:
-		gnutls_free(params_data);
-		gnutls_free(raw_prime.data);
-		gnutls_free(raw_gen.data);
-		gnutls_dh_params_deinit(dh_params);
+	ret = gnutls_dh_get_group(session, &raw_gen, &raw_prime);
+	if (ret) {
+		fprintf(stderr, "gnutls_dh_get_group %d\n", ret);
+		goto out;
 	}
+
+	ret = gnutls_dh_params_init(&dh_params);
+	if (ret) {
+		fprintf(stderr, "gnutls_dh_params_init %d\n", ret);
+		goto out;
+	}
+
+	ret =
+	    gnutls_dh_params_import_raw(dh_params, &raw_prime,
+						&raw_gen);
+	if (ret) {
+		fprintf(stderr, "gnutls_dh_params_import_raw %d\n",
+			ret);
+		goto out;
+	}
+
+	ret = gnutls_dh_params_export_pkcs3(dh_params,
+					    GNUTLS_X509_FMT_PEM,
+					    params_data,
+					    &params_data_size);
+	if (ret != GNUTLS_E_SHORT_MEMORY_BUFFER) {
+		fprintf(stderr,
+			"gnutls_dh_params_export_pkcs3 %d\n", ret);
+		goto out;
+	}
+
+	params_data = gnutls_malloc(params_data_size);
+	if (!params_data) {
+		fprintf(stderr, "gnutls_malloc %d\n", ret);
+		goto out;
+	}
+
+	ret = gnutls_dh_params_export_pkcs3(dh_params,
+					    GNUTLS_X509_FMT_PEM,
+					    params_data,
+					    &params_data_size);
+	if (ret) {
+		fprintf(stderr,
+			"gnutls_dh_params_export_pkcs3-2 %d\n",
+			ret);
+		goto out;
+	}
+
+	log_msg(stdout, " - PKCS#3 format:\n\n%.*s\n",
+	       (int) params_data_size, params_data);
+
+      out:
+	gnutls_free(params_data);
+	gnutls_free(raw_prime.data);
+	gnutls_free(raw_gen.data);
+	gnutls_dh_params_deinit(dh_params);
 #endif
 }
 
-static void print_ecdh_info(gnutls_session_t session, const char *str)
+static void print_ecdh_info(gnutls_session_t session, const char *str, int print)
 {
 	int curve;
 
-	printf("- %sEC Diffie-Hellman parameters\n", str);
+	if (!print)
+		return;
+
+	log_msg(stdout, "- %sEC Diffie-Hellman parameters\n", str);
 
 	curve = gnutls_ecc_curve_get(session);
 
-	printf(" - Using curve: %s\n", gnutls_ecc_curve_get_name(curve));
-	printf(" - Curve size: %d bits\n",
+	log_msg(stdout, " - Using curve: %s\n", gnutls_ecc_curve_get_name(curve));
+	log_msg(stdout, " - Curve size: %d bits\n",
 	       gnutls_ecc_curve_get_size(curve) * 8);
 
 }
@@ -393,13 +406,13 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 	int rc;
 
 	desc = gnutls_session_get_desc(session);
-	printf("- Description: %s\n", desc);
+	log_msg(stdout, "- Description: %s\n", desc);
 	gnutls_free(desc);
 
 	/* print session ID */
 	gnutls_session_get_id(session, session_id, &session_id_size);
 	if (session_id_size > 0) {
-		printf("- Session ID: %s\n",
+		log_msg(stdout, "- Session ID: %s\n",
 		       raw_to_string(session_id, session_id_size));
 	}
 
@@ -412,7 +425,7 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 #ifdef ENABLE_ANON
 	case GNUTLS_CRD_ANON:
 		if (kx == GNUTLS_KX_ANON_ECDH)
-			print_ecdh_info(session, "Anonymous ");
+			print_ecdh_info(session, "Anonymous ", verbose);
 		else
 			print_dh_info(session, "Anonymous ", verbose);
 		break;
@@ -423,7 +436,7 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 		 * side.
 		 */
 		if (gnutls_srp_server_get_username(session) != NULL)
-			printf("- SRP authentication. Connected as '%s'\n",
+			log_msg(stdout, "- SRP authentication. Connected as '%s'\n",
 			       gnutls_srp_server_get_username(session));
 		break;
 #endif
@@ -432,21 +445,21 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 		/* This returns NULL in server side.
 		 */
 		if (gnutls_psk_client_get_hint(session) != NULL)
-			printf("- PSK authentication. PSK hint '%s'\n",
+			log_msg(stdout, "- PSK authentication. PSK hint '%s'\n",
 			       gnutls_psk_client_get_hint(session));
 		/* This returns NULL in client side.
 		 */
 		if (gnutls_psk_server_get_username(session) != NULL)
-			printf("- PSK authentication. Connected as '%s'\n",
+			log_msg(stdout, "- PSK authentication. Connected as '%s'\n",
 			       gnutls_psk_server_get_username(session));
 		if (kx == GNUTLS_KX_DHE_PSK)
 			print_dh_info(session, "Ephemeral ", verbose);
 		if (kx == GNUTLS_KX_ECDHE_PSK)
-			print_ecdh_info(session, "Ephemeral ");
+			print_ecdh_info(session, "Ephemeral ", verbose);
 		break;
 #endif
 	case GNUTLS_CRD_IA:
-		printf("- TLS/IA authentication\n");
+		log_msg(stdout, "- TLS/IA authentication\n");
 		break;
 	case GNUTLS_CRD_CERTIFICATE:
 		{
@@ -457,13 +470,13 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 			/* This fails in client side */
 			if (gnutls_server_name_get
 			    (session, dns, &dns_size, &type, 0) == 0) {
-				printf("- Given server name[%d]: %s\n",
+				log_msg(stdout, "- Given server name[%d]: %s\n",
 				       type, dns);
 			}
 		}
 
 		if ((flags & P_WAIT_FOR_CERT) && gnutls_certificate_get_ours(session) == 0)
-			printf("- No certificate was sent to peer\n");
+			log_msg(stdout, "- No certificate was sent to peer\n");
 
 		if (flags& P_PRINT_CERT)
 			print_cert_info(session, verbose, (flags&P_PRINT_CERT));
@@ -472,65 +485,68 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 			print_dh_info(session, "Ephemeral ", verbose);
 		else if (kx == GNUTLS_KX_ECDHE_RSA
 			 || kx == GNUTLS_KX_ECDHE_ECDSA)
-			print_ecdh_info(session, "Ephemeral ");
+			print_ecdh_info(session, "Ephemeral ", verbose);
 	}
 
-	version = gnutls_protocol_get_version(session);
-	tmp =
-	    SU(gnutls_protocol_get_name(version));
-	printf("- Version: %s\n", tmp);
 
-	if (version < GNUTLS_TLS1_3) {
-		tmp = SU(gnutls_kx_get_name(kx));
-		printf("- Key Exchange: %s\n", tmp);
-	}
-
-	if (gnutls_sign_algorithm_get(session) != GNUTLS_SIGN_UNKNOWN) {
+	if (verbose) {
+		version = gnutls_protocol_get_version(session);
 		tmp =
-		    SU(gnutls_sign_get_name
-		       (gnutls_sign_algorithm_get(session)));
-		printf("- Server Signature: %s\n", tmp);
+		    SU(gnutls_protocol_get_name(version));
+		log_msg(stdout, "- Version: %s\n", tmp);
+
+		if (version < GNUTLS_TLS1_3) {
+			tmp = SU(gnutls_kx_get_name(kx));
+			log_msg(stdout, "- Key Exchange: %s\n", tmp);
+		}
+
+		if (gnutls_sign_algorithm_get(session) != GNUTLS_SIGN_UNKNOWN) {
+			tmp =
+			    SU(gnutls_sign_get_name
+			       (gnutls_sign_algorithm_get(session)));
+			log_msg(stdout, "- Server Signature: %s\n", tmp);
+		}
+
+		if (gnutls_sign_algorithm_get_client(session) !=
+		    GNUTLS_SIGN_UNKNOWN) {
+			tmp =
+			    SU(gnutls_sign_get_name
+			       (gnutls_sign_algorithm_get_client(session)));
+			log_msg(stdout, "- Client Signature: %s\n", tmp);
+		}
+
+		tmp = SU(gnutls_cipher_get_name(gnutls_cipher_get(session)));
+		log_msg(stdout, "- Cipher: %s\n", tmp);
+
+		tmp = SU(gnutls_mac_get_name(gnutls_mac_get(session)));
+		log_msg(stdout, "- MAC: %s\n", tmp);
 	}
 
-	if (gnutls_sign_algorithm_get_client(session) !=
-	    GNUTLS_SIGN_UNKNOWN) {
-		tmp =
-		    SU(gnutls_sign_get_name
-		       (gnutls_sign_algorithm_get_client(session)));
-		printf("- Client Signature: %s\n", tmp);
-	}
-
-	tmp = SU(gnutls_cipher_get_name(gnutls_cipher_get(session)));
-	printf("- Cipher: %s\n", tmp);
-
-	tmp = SU(gnutls_mac_get_name(gnutls_mac_get(session)));
-	printf("- MAC: %s\n", tmp);
-
-	printf("- Options:");
+	log_msg(stdout, "- Options:");
 	if (gnutls_session_ext_master_secret_status(session)!=0)
-		printf(" extended master secret,");
+		log_msg(stdout, " extended master secret,");
 	if (gnutls_safe_renegotiation_status(session)!=0)
-		printf(" safe renegotiation,");
+		log_msg(stdout, " safe renegotiation,");
 	if (gnutls_session_etm_status(session)!=0)
-		printf(" EtM,");
+		log_msg(stdout, " EtM,");
 #ifdef ENABLE_OCSP
 	if (gnutls_ocsp_status_request_is_checked(session, GNUTLS_OCSP_SR_IS_AVAIL)!=0) {
-		printf(" OCSP status request%s,", gnutls_ocsp_status_request_is_checked(session,0)!=0?"":"[ignored]");
+		log_msg(stdout, " OCSP status request%s,", gnutls_ocsp_status_request_is_checked(session,0)!=0?"":"[ignored]");
 	}
 #endif
-	printf("\n");
+	log_msg(stdout, "\n");
 
 #ifdef ENABLE_DTLS_SRTP
 	rc = gnutls_srtp_get_selected_profile(session, &srtp_profile);
 	if (rc == 0)
-		printf("- SRTP profile: %s\n",
+		log_msg(stdout, "- SRTP profile: %s\n",
 		       gnutls_srtp_get_profile_name(srtp_profile));
 #endif
 
 #ifdef ENABLE_ALPN
 	rc = gnutls_alpn_get_selected_protocol(session, &p);
 	if (rc == 0)
-		printf("- Application protocol: %.*s\n", p.size, p.data);
+		log_msg(stdout, "- Application protocol: %.*s\n", p.size, p.data);
 #endif
 
 	if (verbose) {
@@ -545,10 +561,10 @@ int print_info(gnutls_session_t session, int verbose, int flags)
 		else {
 			size_t i;
 
-			printf("- Channel binding 'tls-unique': ");
+			log_msg(stdout, "- Channel binding 'tls-unique': ");
 			for (i = 0; i < cb.size; i++)
-				printf("%02x", cb.data[i]);
-			printf("\n");
+				log_msg(stdout, "%02x", cb.data[i]);
+			log_msg(stdout, "\n");
 			gnutls_free(cb.data);
 		}
 	}
@@ -573,31 +589,16 @@ void print_cert_info2(gnutls_session_t session, int verbose, FILE *out, int prin
 		flag = GNUTLS_CRT_PRINT_COMPACT;
 
 	if (gnutls_certificate_client_get_request_status(session) != 0) {
-		printf("- Server has requested a certificate.\n");
+		log_msg(stdout, "- Server has requested a certificate.\n");
 		print_crt_status = 1;
 	}
 
-	switch (gnutls_certificate_type_get(session)) {
+	switch (gnutls_certificate_type_get2(session, GNUTLS_CTYPE_PEERS)) {
 	case GNUTLS_CRT_X509:
 		print_x509_info(session, out, flag, print_cert, print_crt_status);
 		break;
-	default:
-		break;
-	}
-}
-
-void print_cert_info_compact(gnutls_session_t session)
-{
-	int verbose = 0;
-
-	if (gnutls_certificate_client_get_request_status(session) != 0) {
-		printf("- Server has requested a certificate.\n");
-		verbose = 1;
-	}
-
-	switch (gnutls_certificate_type_get(session)) {
-	case GNUTLS_CRT_X509:
-		print_x509_info_compact(session, verbose);
+	case GNUTLS_CRT_RAWPK:
+		print_rawpk_info(session, out, flag, print_cert, print_crt_status);
 		break;
 	default:
 		break;
@@ -620,7 +621,7 @@ void print_list(const char *priorities, int verbose)
 	const unsigned int *list;
 
 	if (priorities != NULL) {
-		printf("Cipher suites for %s\n", priorities);
+		log_msg(stdout, "Cipher suites for %s\n", priorities);
 
 		ret = gnutls_priority_init(&pcache, priorities, &err);
 		if (ret < 0) {
@@ -646,13 +647,13 @@ void print_list(const char *priorities, int verbose)
 						     NULL, &version);
 
 			if (name != NULL)
-				printf("%-50s\t0x%02x, 0x%02x\t%s\n",
+				log_msg(stdout, "%-50s\t0x%02x, 0x%02x\t%s\n",
 				       name, (unsigned char) id[0],
 				       (unsigned char) id[1],
 				       gnutls_protocol_get_name(version));
 		}
 
-		printf("\n");
+		log_msg(stdout, "\n");
 #if 0
 		{
 			ret =
@@ -660,17 +661,17 @@ void print_list(const char *priorities, int verbose)
 								  &list,
 								  GNUTLS_CTYPE_CLIENT);
 
-			printf("Certificate types: ");
+			log_msg(stdout, "Certificate types: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("CTYPE-%s",
+				log_msg(stdout, "CTYPE-%s",
 				       gnutls_certificate_type_get_name
 				       (list[i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 #endif
@@ -678,64 +679,64 @@ void print_list(const char *priorities, int verbose)
 		{
 			ret = gnutls_priority_protocol_list(pcache, &list);
 
-			printf("Protocols: ");
+			log_msg(stdout, "Protocols: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("VERS-%s",
+				log_msg(stdout, "VERS-%s",
 				       gnutls_protocol_get_name(list[i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 
 		{
 			ret = gnutls_priority_cipher_list(pcache, &list);
 
-			printf("Ciphers: ");
+			log_msg(stdout, "Ciphers: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("%s",
+				log_msg(stdout, "%s",
 				       gnutls_cipher_get_name(list[i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 
 		{
 			ret = gnutls_priority_mac_list(pcache, &list);
 
-			printf("MACs: ");
+			log_msg(stdout, "MACs: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("%s",
+				log_msg(stdout, "%s",
 				       gnutls_mac_get_name(list[i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 
 		{
 			ret = gnutls_priority_kx_list(pcache, &list);
 
-			printf("Key Exchange Algorithms: ");
+			log_msg(stdout, "Key Exchange Algorithms: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("%s",
+				log_msg(stdout, "%s",
 				       gnutls_kx_get_name(list[i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 
@@ -743,33 +744,33 @@ void print_list(const char *priorities, int verbose)
 			ret =
 			    gnutls_priority_group_list(pcache, &list);
 
-			printf("Groups: ");
+			log_msg(stdout, "Groups: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("GROUP-%s",
+				log_msg(stdout, "GROUP-%s",
 				       gnutls_group_get_name(list[i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 
 		{
 			ret = gnutls_priority_sign_list(pcache, &list);
 
-			printf("PK-signatures: ");
+			log_msg(stdout, "PK-signatures: ");
 			if (ret == 0)
-				printf("none\n");
+				log_msg(stdout, "none\n");
 			for (i = 0; i < (unsigned) ret; i++) {
-				printf("SIGN-%s",
+				log_msg(stdout, "SIGN-%s",
 				       gnutls_sign_algorithm_get_name(list
 								      [i]));
 				if (i + 1 != (unsigned) ret)
-					printf(", ");
+					log_msg(stdout, ", ");
 				else
-					printf("\n");
+					log_msg(stdout, "\n");
 			}
 		}
 
@@ -777,99 +778,99 @@ void print_list(const char *priorities, int verbose)
 		return;
 	}
 
-	printf("Cipher suites:\n");
+	log_msg(stdout, "Cipher suites:\n");
 	for (i = 0; (name = gnutls_cipher_suite_info
 		     (i, id, &kx, &cipher, &mac, &version)); i++) {
-		printf("%-50s\t0x%02x, 0x%02x\t%s\n",
+		log_msg(stdout, "%-50s\t0x%02x, 0x%02x\t%s\n",
 		       name,
 		       (unsigned char) id[0], (unsigned char) id[1],
 		       gnutls_protocol_get_name(version));
 		if (verbose)
-			printf
-			    ("\tKey exchange: %s\n\tCipher: %s\n\tMAC: %s\n\n",
+			log_msg
+			    (stdout, "\tKey exchange: %s\n\tCipher: %s\n\tMAC: %s\n\n",
 			     gnutls_kx_get_name(kx),
 			     gnutls_cipher_get_name(cipher),
 			     gnutls_mac_get_name(mac));
 	}
 
-	printf("\n");
+	log_msg(stdout, "\n");
 	{
 		const gnutls_certificate_type_t *p =
 		    gnutls_certificate_type_list();
 
-		printf("Certificate types: ");
+		log_msg(stdout, "Certificate types: ");
 		for (; *p; p++) {
-			printf("CTYPE-%s",
+			log_msg(stdout, "CTYPE-%s",
 			       gnutls_certificate_type_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_protocol_t *p = gnutls_protocol_list();
 
-		printf("Protocols: ");
+		log_msg(stdout, "Protocols: ");
 		for (; *p; p++) {
-			printf("VERS-%s", gnutls_protocol_get_name(*p));
+			log_msg(stdout, "VERS-%s", gnutls_protocol_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_cipher_algorithm_t *p = gnutls_cipher_list();
 
-		printf("Ciphers: ");
+		log_msg(stdout, "Ciphers: ");
 		for (; *p; p++) {
-			printf("%s", gnutls_cipher_get_name(*p));
+			log_msg(stdout, "%s", gnutls_cipher_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_mac_algorithm_t *p = gnutls_mac_list();
 
-		printf("MACs: ");
+		log_msg(stdout, "MACs: ");
 		for (; *p; p++) {
-			printf("%s", gnutls_mac_get_name(*p));
+			log_msg(stdout, "%s", gnutls_mac_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_digest_algorithm_t *p = gnutls_digest_list();
 
-		printf("Digests: ");
+		log_msg(stdout, "Digests: ");
 		for (; *p; p++) {
-			printf("%s", gnutls_digest_get_name(*p));
+			log_msg(stdout, "%s", gnutls_digest_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_kx_algorithm_t *p = gnutls_kx_list();
 
-		printf("Key exchange algorithms: ");
+		log_msg(stdout, "Key exchange algorithms: ");
 		for (; *p; p++) {
-			printf("%s", gnutls_kx_get_name(*p));
+			log_msg(stdout, "%s", gnutls_kx_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
@@ -877,57 +878,94 @@ void print_list(const char *priorities, int verbose)
 		const gnutls_compression_method_t *p =
 		    gnutls_compression_list();
 
-		printf("Compression: ");
+		log_msg(stdout, "Compression: ");
 		for (; *p; p++) {
-			printf("COMP-%s", gnutls_compression_get_name(*p));
+			log_msg(stdout, "COMP-%s", gnutls_compression_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_group_t *p = gnutls_group_list();
 
-		printf("Groups: ");
+		log_msg(stdout, "Groups: ");
 		for (; *p; p++) {
-			printf("GROUP-%s", gnutls_group_get_name(*p));
+			log_msg(stdout, "GROUP-%s", gnutls_group_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_pk_algorithm_t *p = gnutls_pk_list();
 
-		printf("Public Key Systems: ");
+		log_msg(stdout, "Public Key Systems: ");
 		for (; *p; p++) {
-			printf("%s", gnutls_pk_algorithm_get_name(*p));
+			log_msg(stdout, "%s", gnutls_pk_algorithm_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 
 	{
 		const gnutls_sign_algorithm_t *p = gnutls_sign_list();
 
-		printf("PK-signatures: ");
+		log_msg(stdout, "PK-signatures: ");
 		for (; *p; p++) {
-			printf("SIGN-%s",
+			log_msg(stdout, "SIGN-%s",
 			       gnutls_sign_algorithm_get_name(*p));
 			if (*(p + 1))
-				printf(", ");
+				log_msg(stdout, ", ");
 			else
-				printf("\n");
+				log_msg(stdout, "\n");
 		}
 	}
 }
 
+void
+print_key_material(gnutls_session_t session, const char *label, size_t size)
+{
+	gnutls_datum_t bin = { NULL, 0 }, hex = { NULL, 0 };
+	int ret;
+
+	bin.data = gnutls_malloc(size);
+	if (!bin.data) {
+		fprintf(stderr, "Error in gnutls_malloc: %s\n",
+			gnutls_strerror(GNUTLS_E_MEMORY_ERROR));
+		goto out;
+	}
+
+	bin.size = size;
+
+	ret = gnutls_prf_rfc5705(session, strlen(label), label,
+				 0, NULL, size, (char *)bin.data);
+	if (ret < 0) {
+		fprintf(stderr, "Error in gnutls_prf_rfc5705: %s\n",
+			gnutls_strerror(ret));
+		goto out;
+	}
+
+	ret = gnutls_hex_encode2(&bin, &hex);
+	if (ret < 0) {
+		fprintf(stderr, "Error in hex encoding: %s\n",
+			gnutls_strerror(ret));
+		goto out;
+	}
+	log_msg(stdout, "- Key material: %s\n", hex.data);
+	fflush(stdout);
+
+ out:
+	gnutls_free(bin.data);
+	gnutls_free(hex.data);
+}
+
 int check_command(gnutls_session_t session, const char *str, unsigned no_cli_cert)
 {
 	size_t len = strnlen(str, 128);
@@ -984,6 +1022,54 @@ int check_command(gnutls_session_t session, const char *str, unsigned no_cli_cer
 	return 0;
 }
 
+/* error is indicated by returning an empty string */
+void getpass_copy(char *pass, size_t max_pass_size, const char *prompt)
+{
+	char *tmp;
+	size_t len;
+
+	tmp = getpass(prompt);
+	if (tmp == NULL) {
+		pass[0] = 0;
+		return;
+	}
+
+	len = strlen(tmp);
+	if (len >= max_pass_size) {
+		gnutls_memset(tmp, 0, len);
+		pass[0] = 0;
+		return;
+	}
+
+	strcpy(pass, tmp);
+	gnutls_memset(tmp, 0, len);
+
+	return;
+}
+
+/* error is indicated by returning an empty string */
+void getenv_copy(char *str, size_t max_str_size, const char *envvar)
+{
+	char *tmp;
+	size_t len;
+
+	tmp = getenv(envvar);
+	if (tmp == NULL) {
+		str[0] = 0;
+		return;
+	}
+
+	len = strlen(tmp);
+	if (len >= max_str_size) {
+		str[0] = 0;
+		return;
+	}
+
+	strcpy(str, tmp);
+
+	return;
+}
+
 #define MIN(x,y) ((x)<(y))?(x):(y)
 #define MAX_CACHE_TRIES 5
 int
@@ -991,40 +1077,40 @@ pin_callback(void *user, int attempt, const char *token_url,
 	     const char *token_label, unsigned int flags, char *pin,
 	     size_t pin_max)
 {
-	const char *password = NULL;
+	char password[MAX_PIN_LEN] = "";
 	common_info_st *info = user;
 	const char *desc;
 	int cache = MAX_CACHE_TRIES;
 	unsigned len;
 /* allow caching of PIN */
 	static char *cached_url = NULL;
-	static char cached_pin[32] = "";
+	static char cached_pin[MAX_PIN_LEN] = "";
 	const char *env;
 
 	if (flags & GNUTLS_PIN_SO) {
 		env = "GNUTLS_SO_PIN";
 		desc = "security officer";
-		if (info)
-			password = info->so_pin;
+		if (info && info->so_pin)
+			snprintf(password, sizeof(password), "%s", info->so_pin);
 	} else {
 		env = "GNUTLS_PIN";
 		desc = "user";
-		if (info)
-			password = info->pin;
+		if (info && info->pin)
+			snprintf(password, sizeof(password), "%s", info->pin);
 	}
 
 	if (flags & GNUTLS_PIN_FINAL_TRY) {
 		cache = 0;
-		printf("*** This is the final try before locking!\n");
+		log_msg(stdout, "*** This is the final try before locking!\n");
 	}
 	if (flags & GNUTLS_PIN_COUNT_LOW) {
 		cache = 0;
-		printf("*** Only few tries left before locking!\n");
+		log_msg(stdout, "*** Only few tries left before locking!\n");
 	}
 
 	if (flags & GNUTLS_PIN_WRONG) {
 		cache = 0;
-		printf("*** Wrong PIN has been provided!\n");
+		log_msg(stdout, "*** Wrong PIN has been provided!\n");
 	}
 
 	if (cache > 0 && cached_url != NULL) {
@@ -1046,19 +1132,19 @@ pin_callback(void *user, int attempt, const char *token_url,
 		}
 	}
 
-	if (password == NULL) {
-		password = getenv(env);
-		if (password == NULL) /* compatibility */
-			password = getenv("GNUTLS_PIN");
+	if (password[0] == 0) {
+		getenv_copy(password, sizeof(password), env);
+		if (password[0] == 0) /* compatibility */
+			getenv_copy(password, sizeof(password), "GNUTLS_PIN");
 	}
 
-	if (password == NULL && (info == NULL || info->batch == 0 || info->ask_pass != 0)) {
+	if (password[0] == 0 && (info == NULL || info->batch == 0 || info->ask_pass != 0)) {
 		if (token_label && token_label[0] != 0) {
 			fprintf(stderr, "Token '%s' with URL '%s' ", token_label, token_url);
 			fprintf(stderr, "requires %s PIN\n", desc);
-			password = getpass("Enter PIN: ");
+			getpass_copy(password, sizeof(password), "Enter PIN: ");
 		} else {
-			password = getpass("Enter password: ");
+			getpass_copy(password, sizeof(password), "Enter password: ");
 		}
 
 	} else {
@@ -1072,7 +1158,7 @@ pin_callback(void *user, int attempt, const char *token_url,
 		}
 	}
 
-	if (password == NULL || password[0] == 0 || password[0] == '\n') {
+	if (password[0] == 0 || password[0] == '\n') {
 		fprintf(stderr, "No PIN given.\n");
 		if (info != NULL && info->batch != 0) {
 			fprintf(stderr, "note: when operating in batch mode, set the GNUTLS_PIN or GNUTLS_SO_PIN environment variables\n");
@@ -1112,7 +1198,7 @@ token_callback(void *user, const char *label, const unsigned retry)
 		fprintf(stderr, "Could not find token %s\n", label);
 		return -1;
 	}
-	printf("Please insert token '%s' in slot and press enter\n",
+	log_msg(stdout, "Please insert token '%s' in slot and press enter\n",
 	       label);
 	if (fgets(buf, sizeof(buf), stdin) == NULL) {
 		fprintf(stderr, "error reading input\n");
@@ -1146,3 +1232,42 @@ void sockets_init(void)
 	signal(SIGPIPE, SIG_IGN);
 #endif
 }
+
+
+int log_msg(FILE *file, const char *message, ...)
+{
+	va_list args;
+	int rv;
+
+	va_start(args, message);
+
+	rv = vfprintf(logfile ? logfile : file, message, args);
+
+	va_end(args);
+
+	return rv;
+}
+
+void log_set(FILE *file)
+{
+	logfile = file;
+}
+
+/* This is very similar to ctime() but it does not force a newline.
+ */
+char *simple_ctime(const time_t *t, char out[SIMPLE_CTIME_BUF_SIZE])
+{
+	struct tm tm;
+
+	if (localtime_r(t, &tm) == NULL)
+		goto error;
+
+	if (!strftime(out, SIMPLE_CTIME_BUF_SIZE, "%c", &tm))
+		goto error;
+
+	return out;
+
+ error:
+	snprintf(out, SIMPLE_CTIME_BUF_SIZE, "[error]");
+	return out;
+}
